#include "MQPlugin.h"
#include "MQBasePlugin.h"

#include <objbase.h>
#include <gdiplus.h>

#include <unordered_map>

class UVToolsX : public MQStationPlugin
{
public:
	friend UVToolsXWindow;
	friend TextureOutputDialog;

	UVToolsX();
	virtual ~UVToolsX();

	// vOCIDԂB
	virtual void GetPlugInID(DWORD *Product, DWORD *ID);
	// vOCԂB
	virtual const char *GetPlugInName(void);
	// |bvAbvj[ɕ\镶ԂB
	virtual const char *EnumString(void);

	// AvP[V̏
	virtual BOOL Initialize();
	// AvP[V̏I
	virtual void Exit();

	// \E\؂ւv
	virtual BOOL Activate(MQDocument doc, BOOL flag);
	// \E\Ԃ̕ԓ
	virtual BOOL IsActivated(MQDocument doc);

	virtual void OnEndDocument(MQDocument doc);
	virtual void OnDraw(MQDocument doc_, MQScene scene, int width, int height);
	typedef bool (UVToolsX::*ExecuteCallbackProc)(MQDocument doc);
	void Execute(ExecuteCallbackProc proc);

	enum invert_direction_t { INVERT_NO, INVERT_HORIZONTAL, INVERT_VERTICAL };
	enum turn_direction_t { TURN_NO, TURN_UP, TURN_DOWN, TURN_LEFT, TURN_RIGHT };
	enum texturemode_t { TEX_WIRE = 0x1, TEX_VERT = 0x2, TEX_MAT = 0x4, TEX_OBJ = 0x8, TEX_SHADER = 0x10, TEX_CUROBJ = 0x100, TEX_CURMAT = 0x200, CLIPCOLOR = 0x1000, CLIPALPHA = 0x2000 };
	enum filetype_t { FILE_NONE, FILE_BMP, FILE_TGA, FILE_GIF, FILE_PNG, FILE_JPG, FILE_CLIPBOARD };

	// ʂw肵č\_t[eBOI
	BOOL selectFloatingUV(MQDocument doc);
	// _UV_I(IȂƂz)
	BOOL selectUVsame(MQDocument doc);
	// O[vUV_I
	BOOL selectGroupUV(MQDocument doc);

	// UVOBJI
	BOOL selUV2OBJvert(MQDocument doc);
	BOOL selUV2OBJline(MQDocument doc);
	BOOL selUV2OBJface(MQDocument doc);
	// OBJUVI
	BOOL selOBJUV2vert(MQDocument doc);
	BOOL selOBJUV2line(MQDocument doc);
	BOOL selOBJUV2face(MQDocument doc);

	// I𒸓_֐ڒ
	BOOL grabUVvertex(MQDocument doc_, bool faceselectedmode);
	// Iӂ֖ʁAQڒ
	//	BOOL grabUVface(cmq::CMQDocument doc_, bool islandmode = false, bool faceselectedmode = false);
	BOOL grabUVface(MQDocument doc_, bool islandmode = false, bool regular = false, bool notmove = false, bool faceselectedmode = false);
	// KĐڒ
	BOOL grab_regular(MQDocument doc_, bool faceselectedmode);

	// ]
	BOOL invertUVsel(MQDocument doc_, invert_direction_t direction);
	BOOL invertUV(MQDocument doc, float umin, float umax, float vmin, float vmax, invert_direction_t direction);

	// qL
	BOOL turnUVsel(MQDocument doc_, turn_direction_t direction);
	BOOL turnUV(MQDocument doc_, float umin, float umax, float vmin, float vmax, turn_direction_t direction);

	// ]
	BOOL rotate(MQDocument doc_, const MQCoordinate &pivot, float angle);
	// ړ
	BOOL move(MQDocument doc_, const MQCoordinate &delta);
	// gk
	BOOL scale(MQDocument doc_, const MQCoordinate &pivot, float rate);

	// D
	BOOL sewUV(MQDocument doc_, bool faceselectedmode = false, bool grabmode = false);
	// 
	BOOL regularize(MQDocument doc_, bool faceselectedmode);

	// Ώ̉
	BOOL symmetrize(MQDocument doc_, bool widemode);

	// ꒸_
	BOOL gather(MQDocument doc_, bool widemode, float delta);

	// [
	BOOL setMem(MQDocument doc, int index);
	BOOL restoreMem(MQDocument doc, int index);

	// UV_Œ
	bool locksetselected(MQDocument doc_, int index, bool isfix, bool isappend);
	void lockclear(MQDocument doc, int index);
	bool lock(MQDocument doc, int index, bool isfix);

	// UV_͈̔(u,v̍őlAŏl)擾
	void getUVrange(cmq::CMQObject &obj, float &umin, float &umax, float &vmin, float &vmax);

	// eNX`o
	BOOL writefile(MQDocument doc, int width, int height, const std::wstring &filename, DWORD mode, DWORD memory);

	static const int STORESIZE = 4;

protected:
	struct CallbackInfo
	{
		ExecuteCallbackProc proc;
	};

	// R[obNɑ΂
	virtual bool ExecuteCallback(MQDocument doc, void *option);

	class TriangleTree
	{
	public:
		TriangleTree() : m_index(-1), m_done(false) { m_p[0] = m_p[1] = m_p[2] = nullptr; }
		void set(int i, TriangleTree *p) { m_p[i] = p; }
		TriangleTree *get(int i) { return m_p[i]; }
		int m_index;
		bool m_done; // ς݃tO
		TriangleTree *m_p[3];
		cmq::TrianglePoints m_idx;

		static void maketree(std::vector<cmq::TrianglePoints> &tarray, std::vector<TriangleTree> &tree);
		static std::pair<int, int> findneighbor(std::vector<cmq::TrianglePoints> &tarray, int i, int idx1, int idx2);
	};


private:

	// _UV_Xg擾
	std::vector<cmq::flindex_t> getSameUV(cmq::CMQObject &obj, cmq::vindex_t vindex);
	// _UV}bv쐬
	std::unordered_multimap<cmq::vindex_t, cmq::flindex_t> makev2uvmap(cmq::CMQObject &obj);
	// O[vUV_
	BOOL searchUVisland(cmq::CMQObject &obj, std::unordered_set<cmq::findex_t> &foundUV);
	// O[vUV_ċA
	void searchUVisland_recursive(cmq::CMQObject &obj, cmq::findex_t flindex, std::unordered_set<cmq::findex_t> &foundUV, const std::unordered_multimap<cmq::vindex_t, cmq::flindex_t> &v2uvmap, std::unordered_set<cmq::findex_t> &selectedface);

	MQMatrix getUVR(cmq::CMQObject &obj, MQMatrix &r, MQCoordinate &q0, float &rate, cmq::findex_t findexq, int lindex1, int lindex2, const MQCoordinate &pvec);
	void movefaceuv(cmq::CMQObject &obj, cmq::findex_t findex, MQMatrix &r, const MQCoordinate &q0, const cmq::CMQCoordinate &p0, float rate);

	void polygonmove(cmq::CMQObject &obj, cmq::findex_t findex, int lindex1, int lindex2, cmq::CMQCoordinate &p0, MQCoordinate &pvec, float pnorm);
	void func(cmq::CMQObject &obj, TriangleTree *t, std::vector<cmq::vindex_t> &varray, int i, cmq::CMQCoordinate &p0, MQCoordinate &pe, float pveclen, float pnorm, std::vector<cmq::CMQCoordinate> &newpoints);

	void moveface3Duv(cmq::CMQObject &obj, cmq::findex_t findex, MQMatrix &r, const MQPoint &q0, const cmq::CMQCoordinate &p0, float rate);
	MQMatrix get3DR(cmq::CMQObject &obj, MQMatrix &r, MQPoint &q0, float &rate, cmq::findex_t findexq, int lindex1, int lindex2, const MQCoordinate &pvec, float pnorm);

	void getFaceFrom2Vert(cmq::CMQObject &obj, std::vector<std::pair<cmq::flindex_t, cmq::flindex_t>> &uvlist, cmq::vindex_t vindex1, cmq::vindex_t vindex2);

	// eNX`o͗p
	void writedata(MQDocument doc_, long width, long height, const std::wstring &filename, DWORD mode, DWORD memory, filetype_t filetype);
	BOOL writewire(cmq::CMQDocument &doc, Gdiplus::Bitmap *bitmap, long width, long height, bool curobj, bool curmat, DWORD mode, DWORD memory);
	BOOL writevertex(cmq::CMQDocument &doc, Gdiplus::Bitmap *bitmap, long width, long height, bool curobj, bool curmat, DWORD mode, DWORD memory);
	BOOL writemat(cmq::CMQDocument &doc, Gdiplus::Bitmap *bitmap, long width, long height, bool curobj, bool curmat, DWORD mode, DWORD memory, bool ismat);
	// AevZ
	bool calcshade(cmq::CMQDocument &doc, cmq::CMQObject &obj, std::unordered_map<cmq::flindex_t, MQColor> &shademap);
	// ɋLꂽʂ`FbN
	bool memorycheck(cmq::CMQObject &obj, DWORD memory, cmq::findex_t findex);

	UVToolsXWindow *m_Window;
	bool m_bActivate;

	std::unordered_map<UINT, std::unordered_set<cmq::flindex_t>> m_store[STORESIZE];
	std::unordered_map<UINT, std::unordered_map<cmq::flindex_t, MQCoordinate>> m_lock[STORESIZE];
	bool m_isfix[STORESIZE]; // Œ蒆tO

	std::wstring m_filename;
	DWORD m_texturemode;
	int m_texturewidth;
	int m_textureheight;
};

// v}EXJ[\Ǘ
class CursorKeeper
{
public:
	CursorKeeper(LPCWSTR lpname = IDC_WAIT) { setCursor(lpname); }
	~CursorKeeper() { setCursor(IDC_ARROW); }

private:
	void setCursor(LPCWSTR lpname)
	{
		HCURSOR hcur = (HCURSOR)LoadImage(NULL, lpname, IMAGE_CURSOR, 0, 0, LR_SHARED);
		SetCursor(hcur);
	}
};

#pragma pack(push, 1)

struct TGAFILEHEADER_COLORMAPSPEC
{
	WORD colormapindex;
	WORD colormaplength;
	BYTE colormapsize;
};

struct TGAFILEHEADER_IMAGESPEC
{
	WORD imageoriginX;
	WORD imageoriginY;
	WORD imagewidth;
	WORD imageheight;
	BYTE bitperpixel;
	BYTE descripter;
};

struct TGAFILEHEADER
{
	BYTE id;
	BYTE colormaptype;
	BYTE imagetype;
	TGAFILEHEADER_COLORMAPSPEC colormap;
	TGAFILEHEADER_IMAGESPEC imagemap;
};

struct TGAFILEFOOTER
{
	DWORD extensionoffset;
	DWORD developeroffset;
	char signature[18];
};

#pragma pack(pop)
